/*
  CPP_CONTEST=2018Junio
  CPP_PROBLEM=E
  CPP_LANG=CPP+OPENMP+MPI
  CPP_PROCESSES_PER_NODE=calisto 2
*/

#include <stdlib.h>
#include <mpi.h>
#include <omp.h>

extern void escribir(double *,int,int);

void copiar(int n,double *vd,int ldvd,double *vo,int ldvo) 
//se copia el vector vo en el vd
//como los datos pueden no estar contiguos se accede a ellos con los ld
{
  int i;
  for(i=0;i < n;i++)
    vd[i*ldvd]=vo[i*ldvo];
}

void ordenar(int n,double *m)
{
  int i,j;
  double t;
  for(i=0;i < n;i++)
    for(j=i+1;j < n;j++)
    {
      if(m[i] > m[j])
      {
        t=m[i];
        m[i]=m[j];
        m[j]=t;
      }
    }
}

void ordenarfilas(int n,double *A,double *F)
{
  int i;
#pragma omp parallel for private(i)
  for(i=0;i < n;i++)
  { 
    copiar(n,&F[i*n],1,&A[i*n],1); //se copia la fila de A en la correspondiente de F
    ordenar(n,&F[i*n]); //y se ordena la fila en F
  }
}

void trasponer(int n,double *m)
{
  int i,j;
  double t;
  for(i=0;i < n;i++)
    for(j=i+1;j < n;j++)
    {
       t=m[i*n+j];
       m[i*n+j]=m[j*n+i];
       m[j*n+i]=t;
    }
}

void ordenarcolumnas(int n,double *A,double *C)
{
  int i;
#pragma omp parallel for private(i)
  for(i=0;i < n;i++)
  { 
    copiar(n,&C[i*n],1,&A[i],n); //se copia la columna de A en la fila correspondiente de C
    ordenar(n,&C[i*n]); //y se ordena la fila en C
  }
  trasponer(n,C); //se pasan las filas de C, que tienen las columnas ordenadas de A, a las columnas correspondientes
}

void sumar(int n,double *vd,double *vo1,double *vo2)
{
  int i;
#pragma omp parallel for private(i)
  for(i=0;i < n*n;i++)
    vd[i]=vo1[i]+vo2[i];
}

void  sec(int n,double *A,int nodo,int np)
{
  MPI_Status *status;
  MPI_Bcast(&n,1,MPI_INT,0,MPI_COMM_WORLD);
  if(nodo==1)
    A=new double[n*n];   
  double *B=new double[n*n];
  MPI_Bcast(A,n*n,MPI_DOUBLE,0,MPI_COMM_WORLD);

  if(nodo==0)
  {
    ordenarfilas(n,A,A);
#ifdef DEBUG
  escribir(B,n,n);
#endif
    MPI_Recv(B,n*n,MPI_DOUBLE,1,40,MPI_COMM_WORLD,status);
    sumar(n,A,A,B);
#ifdef DEBUG
  escribir(A,n,n);
#endif
  }
  else if(nodo==1)
  {
     ordenarcolumnas(n,A,B);
#ifdef DEBUG
  escribir(C,n,n);
#endif
     MPI_Send(B,n*n,MPI_DOUBLE,0,40,MPI_COMM_WORLD);
     delete[] A;
  }
  delete[] B;
}

